package cn.com.qmth.examcloud.ws.starter.core;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.support.SpringContextHolder;
import com.google.gson.JsonElement;
import com.google.gson.JsonParser;
import com.google.gson.JsonSyntaxException;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.logging.log4j.ThreadContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import javax.websocket.*;
import javax.websocket.server.PathParam;
import javax.websocket.server.ServerEndpoint;
import java.io.EOFException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicLong;

/**
 * websocket ServerEndpoint
 *
 * @author WANGWEI
 * @date 2019年11月27日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
@ServerEndpoint("/api/ws/{path}")
@Component
public class WebSocketServerEndpoint {

    private static final Logger LOG = LoggerFactory.getLogger(WebSocketServerEndpoint.class);

    private static RedisClient redisClient;

    private static AtomicLong counter = new AtomicLong(0L);

    private String sessionId;

    private static RedisClient getRedisClient() {
        if (null != redisClient) {
            return redisClient;
        }
        redisClient = SpringContextHolder.getBean(RedisClient.class);
        return redisClient;
    }

    @OnOpen
    public void onOpen(Session session, @PathParam("path") String path) {
        long amount = counter.incrementAndGet();
        this.sessionId = UUID.randomUUID().toString().replace("-", "");
        ThreadContext.put("TRACE_ID", sessionId);

        String key = getRequestParameter(session, "key");
        String token = getRequestParameter(session, "token");
        if (StringUtils.isBlank(key)) {
            LOG.warn("[onOpen] refused... path={}, user key is empty", path);
            IOUtils.closeQuietly(session);
            return;
        }
        if (StringUtils.isBlank(token)) {
            LOG.warn("[onOpen] refused... path={}, user token is empty", path);
            IOUtils.closeQuietly(session);
            return;
        }

        MessageOut out = new MessageOut(path, sessionId);
        out.setEventId("open");

        Method method = MessageHandlerHolder.getMethod(path);
        if (null == method) {
            LOG.warn("[onOpen] refused... user={} path={}, 404 not found", key, path);

            out.setStatus("404", "not found");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        User user = getRedisClient().get(key, User.class);
        if (null == user) {
            LOG.warn("[onOpen] refused... user={} path={}, 403 no login", key, path);

            out.setStatus("403", "no login");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        if (!token.equals(user.getToken())) {
            LOG.warn("[onOpen] refused... user={} path={}, 403 token invalid", key, path);

            out.setStatus("403", "token invalid");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        ThreadContext.put("CALLER", user.getKey());

        if (LOG.isInfoEnabled()) {
            LOG.info("[onOpen] into... user={} path={} wsSessionSize={}", key, path, amount);
        }

        WebSocketHelper.sendText(session, path, out);

        WebSocketHelper.setSession(path, user, session, sessionId);
    }

    /**
     * 获取请求参数
     *
     * @param session
     * @param key
     * @return
     * @author WANGWEI
     */
    private String getRequestParameter(Session session, String key) {
        Map<String, List<String>> params = session.getRequestParameterMap();
        List<String> list = params.get(key);

        if (CollectionUtils.isEmpty(list)) {
            return null;
        }

        return StringUtils.join(list, ",");
    }

    @OnClose
    public void onClose(Session session, @PathParam("path") String path) {
        long amount = counter.decrementAndGet();

        String user = "unknown";
        SessionInfo sessionInfo = WebSocketHelper.getSessionInfo(session);
        if (null != sessionInfo) {
            user = sessionInfo.getUser().getKey();
            ThreadContext.put("TRACE_ID", sessionInfo.getSessionId());
            ThreadContext.put("CALLER", user);
        } else {
            // 临时会话
            ThreadContext.put("TRACE_ID", this.sessionId);
        }

        if (LOG.isInfoEnabled()) {
            LOG.info("[onClose] user={} path={} wsSessionSize={}", user, path, amount);
        }

        WebSocketHelper.closeSession(session);
    }

    @OnMessage
    public void onMessage(Session session, @PathParam("path") String path, String message) {
        SessionInfo sessionInfo = WebSocketHelper.getSessionInfo(session);
        String userKey = sessionInfo.getUser().getKey();
        ThreadContext.put("TRACE_ID", sessionInfo.getSessionId());
        ThreadContext.put("CALLER", userKey);

        MessageOut out = new MessageOut(path, sessionInfo.getSessionId());
        try {
            JsonElement jsonEl = JsonParser.parseString(message);
            message = JsonUtil.toJson(jsonEl);
        } catch (JsonSyntaxException e) {
            LOG.warn("[onMessage] fail... user={} path={}, 500 message invalid: {}", userKey, path, message);

            out.setStatus("500", "message invalid");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        User user = getRedisClient().get(userKey, User.class);
        if (null == user) {
            LOG.warn("[onMessage] fail... user={} path={}, 403 token invalid", userKey, path);

            out.setStatus("403", "no login");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        if (!sessionInfo.getUser().getToken().equals(user.getToken())) {
            LOG.warn("[onMessage] fail... user={} path={}, 403 token invalid", userKey, path);

            out.setStatus("403", "token invalid");
            WebSocketHelper.sendText(session, path, out);
            IOUtils.closeQuietly(session);
            return;
        }

        boolean hasErr = false;
        Object result = null;
        try {
            Method method = MessageHandlerHolder.getMethod(path);
            Object bean = MessageHandlerHolder.getBean(path);

            Annotation[][] an2 = method.getParameterAnnotations();

            boolean hasMessageParam = false;
            int messageParamIndex = 0;
            OUTER:
            for (int i = 0; i < an2.length; i++) {
                Annotation[] an1 = an2[i];
                for (Annotation an : an1) {
                    if (an.annotationType().equals(Message.class)) {
                        hasMessageParam = true;
                        messageParamIndex = i;
                        break OUTER;
                    }
                }
            }

            Class<?>[] parameterTypes = method.getParameterTypes();
            Object[] args = new Object[parameterTypes.length];
            for (int i = 0; i < parameterTypes.length; i++) {
                Class<?> curType = parameterTypes[i];
                if (hasMessageParam && i == messageParamIndex) {
                    try {
                        args[i] = JsonUtil.fromJson(message, curType);
                    } catch (Exception e) {
                        throw new StatusException("500", "message parse fail...");
                    }
                    continue;
                }
                if (curType.equals(User.class)) {
                    args[i] = user;
                    continue;
                }
            }

            result = method.invoke(bean, args);

            out.setContent(result);
        } catch (StatusException e) {
            LOG.error("[onMessage] error... user={} path={}", userKey, path, e);
            out.setStatus(e.getCode(), e.getDesc());
            hasErr = true;
        } catch (Exception e) {
            LOG.error("[onMessage] error... user={} path={}", userKey, path, e);
            out.setStatus("500", "系统异常");
            hasErr = true;
        }

        if (LOG.isInfoEnabled() && !hasErr) {
            if (result != null) {
                LOG.info("[onMessage] user={} path={} content={}", userKey, path, result);
            } else {
                LOG.info("[onMessage] user={} path={}", userKey, path);
            }
        }

        WebSocketHelper.sendText(session, path, out);
    }

    @OnError
    public void onError(Session session, @PathParam("path") String path, Throwable err) {
        String user = "unknown";
        SessionInfo sessionInfo = WebSocketHelper.getSessionInfo(session);
        if (null != sessionInfo) {
            user = sessionInfo.getUser().getKey();
            ThreadContext.put("TRACE_ID", sessionInfo.getSessionId());
            ThreadContext.put("CALLER", user);
        } else {
            // 临时会话
            ThreadContext.put("TRACE_ID", this.sessionId);
        }

        if (err != null) {
            if (err instanceof EOFException) {
                // WebSocket 默认60秒内无心跳会关闭连接
                LOG.error("[onError] user={} path={}, cause = connection closed, EOFException...", user, path);
            } else {
                LOG.error("[onError] user={} path={}", user, path, err);
            }
        } else {
            LOG.error("[onError] user={} path={}, no err msg...", user, path);
        }

        WebSocketHelper.closeSession(session);
    }

}
